#include "fsm.h"

FSM::FSM ()
{
	alphabetSize = 0;
	bInit = false;
}

FSM::~FSM ()
{
	deAllocFSM();
}

FSM::FSM (const FSM& r) : charSetMap (r.charSetMap)
{
	alphabetSize = r.alphabetSize;
	finalStatus = r.finalStatus;
	
	for (int i = 0; i < r.table.size(); i++)
	{
		Word* newEntry = new Word [alphabetSize];
		if (!newEntry)
			return;

		memcpy ((void*)newEntry, (const void*)r.table[i], alphabetSize*sizeof(Word));
		table.push_back(newEntry);
	}

	bInit = true;
}

void FSM::deAllocFSM ()
{
	for (int i = 0; i < table.size(); i++)
		if (table[i] != NULL) delete [] table[i];

	table.clear();
	finalStatus.clear();
	alphabetSize = 0;
	charSetMap.Clear ();
	bInit = false;
	return;
}

// builds FSM based on given list of patterns
bool FSM::allocFSM (vector <wstring> patterns)
{
	// deallocate prev FSM (if any)
	deAllocFSM ();

	// alloc new one basedon given list of pattrens
	
	// build char set map for all chars in patterns' list
	int i;
	for (i = 0; i < patterns.size(); i++)
		for (int j = 0; j < patterns[i].length(); j++)
			// check if the char already exists; if not add it in list
			if (charSetMap.GetIndex(patterns[i][j]) == UNDEF_CHAR)
				charSetMap.Insert(patterns[i][j]);

	alphabetSize = charSetMap.GetTotalNumOfChars ();

	// insert patterns in FSM
	// create start state (0)
	Word* newEntry = new Word [alphabetSize];
	if (!newEntry)
		return false;

	memset ((void*)newEntry, 0, alphabetSize*sizeof(Word));
	table.push_back(newEntry);
	finalStatus.push_back(NON_FINAL); // set non-final state

	Word currState, nextState;
	for (i = 0; i < patterns.size(); i++)
	{
		currState = 0;	// set to start state

		for (int j = 0; j < patterns[i].length(); j++)
		{
			Word charIndex = charSetMap.GetIndex(patterns[i][j]);

			if (charIndex == UNDEF_CHAR)
				return false;	// unexpected unknown char found

			nextState = table[currState][charIndex];
			if (nextState == 0)	// no next state exists for curr char at curr state
			{	// create new state and add its link
				newEntry = new Word [alphabetSize];
				if (!newEntry)
					return false;

				memset ((void*)newEntry, 0, alphabetSize*sizeof(Word));
				table.push_back(newEntry);
				finalStatus.push_back(NON_FINAL); // set non-final state
				nextState = table.size() - 1;

				table[currState][charIndex] = nextState;
			}

			currState = nextState;
		}

		// set the final state for curr pattern
		finalStatus[currState] = i;	// saves index of curr pattern
	}

	bInit = true;
	return true;
}

POccurrence FSM::find (wstring text, unsigned int sIndex)
{
	POccurrence pOccurrence = {-1, 0xFFFF};
	if (!bInit)	// check whether FSM is defined
		return pOccurrence;
	
	Word currState, nextState;
	int i = sIndex;
	for (i = sIndex; i < text.size() && pOccurrence.location == -1; i++)
	{
		currState = 0;
		for (int j = i; j < text.size(); j++)
		{
			Word charIndex = charSetMap.GetIndex(text[j]);
			if (charIndex == UNDEF_CHAR)
				break;	// unknown char found

			nextState = table[currState][charIndex];
			if (nextState == 0)	// no next state exists for curr char at curr state
				break;

			currState = nextState;
			if (finalStatus[currState] != NON_FINAL) // if curr state is final state
				// save curr states index if has high precedence (low in number)
				if (finalStatus[currState] < pOccurrence.patternID)
				{
					pOccurrence.patternID = finalStatus[currState];
					pOccurrence.location = i;
				}
		}
	}

	return pOccurrence;
}

bool FSM::findWhole (wstring text)
{
	if (!bInit)	// check whether FSM is defined
		return false;
	
	Word currState, nextState;

	currState = 0;
	for (int j = 0; j < text.size(); j++)
	{
		Word charIndex = charSetMap.GetIndex(text[j]);
		if (charIndex == UNDEF_CHAR)
			return false;	// unknown char found

		nextState = table[currState][charIndex];
		if (nextState == 0)	// no next state exists for curr char at curr state
			return false;

		currState = nextState;
	}

	if (finalStatus[currState] != NON_FINAL) // if curr state is final state
		return true;

	return false;
}
